/***********************************************************************
 * File: byteorder.c
 * 	(Extracted from
 *	   "$Id: byteorder.c,v 1.2 2009/05/21 13:31:45 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source file implements `byteorder' API-s.
 *
 * Copyright (C) 2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "byteorder.h"
#ifdef S_SPLINT_S /* { */
	/* for `splint' parse error in "net/if.h" included from
	 * "arpa/inet.h"
	 */
	typedef char	*caddr_t;
#endif /* } defined(S_SPLIT_S) */
#include <arpa/inet.h> /* for `ntohl, ntohs, htonl, htons' */


/* transfer the octets (unsigned 8bits integer) sequence */
static char *conv_chars(Byteorder_dir_e	direction
						, char			*netp
						, char			*hostp
						, size_t		occurrence
						)
{
	char	*src, *dst;

	/* set addresses of source and destination. */
	if (direction == Byteorder_dir_e_ntoh) {
		/* network-byteorder to host's one */
		src = netp; dst = hostp;
	}
	else {	/* host-byteorder to network's one */
		src = hostp; dst = netp;
	}

	/* test: `src' is not the same address as `dst'? */
	if (src != dst) {
		/* test: `dst' buffer doesn't overlap `src'? */
		if (dst < src || (src + occurrence) <= dst) {
			/* transfer buffers (as is) */
			memcpy(dst, src, occurrence);
		}
		else {	/* transfer overlapped buffers */
			for (src += (occurrence - 1), dst += (occurrence - 1)
			; occurrence-- > 0
			; --src, --dst
			) {	*dst	= *src;	}
		}
	}

	return	dst; /* return address to the destination buffer */
}


/* Macro: transfer integers  [ */
#define Conv_integers(TYPE,SRC,DST,CONVMACRO) {		\
	TYPE	val, new_val;							\
	while (occurrence-- > 0) {						\
		memcpy(&val, SRC, sizeof(val));				\
		new_val	= CONVMACRO(val);					\
		memcpy(DST, &new_val, sizeof(new_val));		\
		SRC += sizeof(val); DST += sizeof(val);		\
	}												\
}

#define Converter(TYPE,NTOH,HTON) {					\
	char	*result;								\
	if (direction == Byteorder_dir_e_ntoh) {		\
		Conv_integers(TYPE, netp, hostp, NTOH);		\
		result	= hostp;							\
	}												\
	else {											\
		Conv_integers(TYPE, hostp, netp, HTON);		\
		result	= netp;								\
	}												\
	return	result;									\
}

/* transfer unsigned 32bits integer sequence */
static char *conv_longs(Byteorder_dir_e	direction
						, char			*netp
						, char			*hostp
						, size_t		occurrence
						)
{
	Converter(uint32_t, ntohl, htonl);
}

/* transfer unsigned 16bits integer sequence */
static char *conv_shorts(Byteorder_dir_e	direction
						, char				*netp
						, char				*hostp
						, size_t			occurrence
						)
{
	Converter(uint16_t, ntohs, htons);
}

#undef Converter
#undef Conv_integers /* ] Macro: transfer integers */


/* Byteorder converting for the sequence of 32bits integers
 * Arguments;
 *	o `direction' indicates converting directions;
 *		(1) Byteorder_dir_e_ntoh: network-wide to host-native
 *		(2) Byteorder_dir_e_hton: host-native to network-wide
 *	o `netp' points to the sequence of network-byteorder 32bits
 *	 integers.
 *	o `hostp' points to the sequence of host-byteorder 32bits
 *	 integers.
 *	o `occurrence' indicates the number of integers sequences.
 * Returns;
 *	NULL    : several errors occurred.
 *	non-NULL: return normally
 */
char	*byteorder_longs(Byteorder_dir_e	direction
						, char				*netp
						, char				*hostp
						, size_t			occurrence
						)
{
	/* test: invalid arguments? */
	if (((((direction < Byteorder_dir_e_ntoh)
	|| direction >= Byteorder_dir_e_N)
	|| netp == NULL) || hostp == NULL) || occurrence == 0
	) {	/* ERROR: invalid arguments */
		errno = EINVAL; return NULL; /* return error */
	}

	/* return address to the destination buffer */
	return	conv_longs(direction, netp, hostp, occurrence);
}

/* Byteorder converting for the sequence of 16bits integers
 * Arguments;
 *	o `direction' indicates converting directions;
 *		(1) Byteorder_dir_e_ntoh: network-wide to host-native
 *		(2) Byteorder_dir_e_hton: host-native to network-wide
 *	o `netp' points to the sequence of network-byteorder 16bits
 *	 integers.
 *	o `hostp' points to the sequence of host-byteorder 16bits
 *	 integers.
 *	o `occurrence' indicates the number of integers sequences.
 * Returns;
 *	NULL    : several errors occurred.
 *	non-NULL: return normally
 */
char	*byteorder_shorts(Byteorder_dir_e	direction
							, char			*netp
							, char			*hostp
							, size_t		occurrence
							)
{
	/* test: invalid arguments? */
	if (((((direction < Byteorder_dir_e_ntoh)
	|| direction >= Byteorder_dir_e_N)
	|| netp == NULL) || hostp == NULL) || occurrence == 0
	) {	/* ERROR: invalid arguments */
		errno = EINVAL; return NULL; /* return error */
	}

	/* return address to the destination buffer */
	return	conv_shorts(direction, netp, hostp, occurrence);
}

/* Byteorder converting for the sequence of 8bits integers
 * Arguments;
 *	o `direction' indicates converting directions;
 *		(1) Byteorder_dir_e_ntoh: network-wide to host-native
 *		(2) Byteorder_dir_e_hton: host-native to network-wide
 *	o `netp' points to the sequence of network-byteorder 8bits
 *	 integers.
 *	o `hostp' points to the sequence of host-byteorder 8bits
 *	 integers.
 *	o `occurrence' indicates the number of integers sequences.
 * Returns;
 *	NULL    : several errors occurred.
 *	non-NULL: return normally
 */
char	*byteorder_chars(Byteorder_dir_e	direction
						, char			*netp
						, char			*hostp
						, size_t		occurrence
						)
{
	/* test: invalid arguments? */
	if (((((direction < Byteorder_dir_e_ntoh)
	|| direction >= Byteorder_dir_e_N)
	|| netp == NULL) || hostp == NULL) || occurrence == 0
	) {	/* ERROR: invalid arguments */
		errno = EINVAL; return NULL; /* return error */
	}

	/* return address to the destination buffer */
	return	conv_chars(direction, netp, hostp, occurrence);
}

/* Byteorder converting for the differ buffers.
 * Arguments;
 *	o `direction' indicates converting directions;
 *		(1) Byteorder_dir_e_ntoh: network-wide to host-native
 *		(2) Byteorder_dir_e_hton: host-native to network-wide
 *	o `netp' points to the network-byteorder buffer.
 *	o `net_rec_sz' indicates the size of `netp' records.
 *	o `layouts_n' indicates the number of elements of `layouts'
 *	 array.
 *	o `layouts' points to the array of layout of the converting
 *	 fields.
 *	o `hostp' points to the host-byteorder buffer.
 *	o `host_rec_sz' indicates the size of `hostp' records.
 * Returns;
 *	NULL    : several errors occurred.
 *	non-NULL: return normally
 */
char	*byteorder(Byteorder_dir_e			direction
					, char					*netp
					, size_t				net_rec_sz
					, size_t				layouts_n
					, Byteorder_layout_t	*layouts
					, char					*hostp
					, size_t				host_rec_sz
					)
{
	size_t	i;

	/* test: invalid arguments? */
	if ((((((((direction < Byteorder_dir_e_ntoh)
	|| direction >= Byteorder_dir_e_N)
	|| netp == NULL)	|| net_rec_sz == 0)
	|| layouts_n == 0)	|| layouts == NULL)
	|| hostp == NULL)	|| host_rec_sz == 0
	) {	/* ERROR: invalid arguments */
		errno = EINVAL; return NULL; /* return error */
	}
	for (i = 0; i < layouts_n; ++i) {
		if (((((layouts[i].type < Byteorder_type_e_char)
		|| layouts[i].type >= Byteorder_type_e_N)
		|| layouts[i].occurrence == 0)
		|| (layouts[i].net_fld_offset
				+ (Byteorder_type_to_size(layouts[i].type)
						* layouts[i].occurrence
				)) > net_rec_sz)
		|| (layouts[i].host_fld_offset
				+ (Byteorder_type_to_size(layouts[i].type)
					* layouts[i].occurrence
				)) > host_rec_sz)
		{	/* ERROR: invalid `layouts' argument */
			errno = EINVAL; return NULL; /* return error */
	}	}

	/* transfer the each fields... */
	for (i = 0; i < layouts_n; ++layouts, ++i) {
		switch (layouts->type) {
		case Byteorder_type_e_long: /* unsigned 32bits */
			(void)conv_longs(direction
							, (netp + layouts->net_fld_offset)
							, (hostp + layouts->host_fld_offset)
							, (layouts->occurrence)
							);
			break;

		case Byteorder_type_e_short: /* unsigned 16bits */
			(void)conv_shorts(direction
							, (netp + layouts->net_fld_offset)
							, (hostp + layouts->host_fld_offset)
							, (layouts->occurrence)
							);
			break;

		case Byteorder_type_e_char: /* octets (8bits) */
		default:
			(void)conv_chars(direction
							, (netp + layouts->net_fld_offset)
							, (hostp + layouts->host_fld_offset)
							, (layouts->occurrence)
							);
			break;
		}
	}

	/* return address to the destination buffer */
	return	(direction == Byteorder_dir_e_ntoh)? hostp: netp;
}
