/***********************************************************************
 * File: conf_file_loader.h
 * 	(Extracted from
 *	   "$Id: conf_file_loader.h,v 1.2 2009/05/21 12:57:26 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  	This C header defines the types and macros for "load the
 *  configuration files" API-s.
 *
 * Copyright (C) 2005-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#ifndef _LOAD_CONF_FILE_H_ /* { */
#define _LOAD_CONF_FILE_H_
#ifdef __cplusplus
extern "C" {
#endif

/* include the system definitions */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <memory.h>
#include <errno.h>


/* type definitions for loading the configuration files */
typedef struct conf_row_t { /* for the row informations */
	char	*key;	       /* pointing to the key string */
	size_t	vals_n;	       /* number of the values */
	size_t	vals_alloc_sz; /* allocated size of the values vector */
	char	**vals;	       /* pointing to the values vector */
} Conf_row_t;

typedef struct configuration_t { /* for the configuration info */
	char		*top_addr; /* pointing to the file contents */
	size_t		file_alloc_sz;	/* allocated size of contents */
	size_t		rows_n;		/* number of the rows */
	size_t		rows_alloc_sz;	/* allocated size of rows */
	Conf_row_t	rows[1];	/* pointing to the rows info */
} Configuration_t;


/* definitions of the comments characters */
#define	JAVA_PROPERTY_FILE_COMMENT_CHARS "#!/*"
				/* for Java property styles */
#define DOS_INI_FILE_COMMENT_CHARS	 "[;/"
				/* for DOS INI-file styles */


/*
 * This API loads the configuration file, that's style is DOS's
 * INI-files or Java's property-files, and the builds the
 * Configuration_t structure.
 *
 * arguments:
 *	path is pointing to the path of the configuration file.
 *	comment_chars is pointing to the comment characters.
 *	val_delimit indicates the delimit character of the values.
 *
 * returns:
 *	non-NULL: the Configuration_t structure that is loaded the file.
 *	NULL: several failure is happened. errno indicates the details.
 *		errno : means...
 *		EINVAL: illegal arguments
 *		EACCES: the file isn't readable access mode
 *		EIO:	failed to read the file to allocated area
 *		ENOMEM:	insufficient memory
 */
extern Configuration_t *load_conf_file(char *path
										, char *comment_chars
										, char val_delimit
										);


/*
 * This API get the address of Conf_row_t slot in the Configure_t,
 * that's key is matched the specified key string.
 *
 * arguments:
 *	conf is pointing the returns of the load_conf_file() API.
 *	key is pointing to the key string.
 *
 * returns:
 *	non-NULL: pointing the Conf_row_t slot that is matched the key
 *		string.
 *	NULL: the Conf_row_t is empty in the conf, or illegal arguments
 *		errno : means...
 *		zero  : the Conf_row_t slot of conf is empty.
 *		EINAVL: illegal arguments
 */
extern Conf_row_t *get_row_matched_key(Configuration_t *conf, char *key);


/*
 * This API find the value slot in the value vector, that is pointed
 * from the Conf_row_t in conf that is matched with key string. And
 * outputs the index of the values vector.
 *
 * arguments:
 *	conf is the returns of load_conf_file() API.
 *	key is pointing to the key string.
 *	value is pointing the finding value string.
 *	val_idx outputs the index of the values vector.
 *
 * returns:
 *	non-NULL: address of the found value string in the vector
 *	NULL: not-found the value in the vector, or illegal arguments
 *		errno : means...
 *		zero  : not found the value in the values vector
 *		EINVAL: illegal arguments
 */
extern char *find_value_matched_key(Configuration_t *conf
									, char *key
									, char *value
									, size_t	*val_idx
									);


/*
 * This API release the allocated memories for Configuration_t structure
 *
 * arguments:
 *	conf is the returns of load_conf_file() API.
 *
 * returns:
 *	nothing
 */
extern void destroy_conf(Configuration_t *conf);

#ifdef __cplusplus
}
#endif
#endif /* } !defined(_LOAD_CONF_FILE_H_) */
